# oscillode
# Copyright (C) 2024 oscillode Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# blender_maxwell
# Copyright (C) 2024 blender_maxwell Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `ModeSolverNode`."""

import enum
import typing as typ

import bpy
import sympy as sp
import tidy3d as td

from blender_maxwell.utils import bl_cache, logger
from blender_maxwell.utils import sympy_extra as spux
from blender_maxwell.utils.frozendict import frozendict

from ... import contracts as ct
from ... import sockets
from .. import base, events

log = logger.get(__name__)

FK = ct.FlowKind
FS = ct.FlowSignal
MT = spux.MathType
PT = spux.PhysicalType


class ModePol(enum.StrEnum):
	"""Polarization filter with which to arrange mode indices with."""

	Arbitrary = enum.auto()
	TE = enum.auto()
	TM = enum.auto()

	@property
	def tidy3d_pol(self) -> str | None:
		MP = ModePol
		return {
			MP.Arbitrary: None,
			MP.TE: 'te',
			MP.TM: 'tm',
		}[self]

	####################
	# - UI
	####################
	@staticmethod
	def to_name(value: typ.Self) -> str:
		"""A human-readable UI-oriented name for a physical type."""
		MP = ModePol
		return {
			MP.Arbitrary: 'Arbitrary',
			MP.TE: 'TE',
			MP.TM: 'TM',
		}[value]

	@staticmethod
	def to_icon(_: typ.Self) -> str:
		"""No icons."""
		return ''


class RunModalSolve(bpy.types.Operator):
	"""Run a modal solver on data provided to a `ModeSolverNode`."""

	bl_idname = ct.OperatorType.NodeSolveModes
	bl_label = 'Solve Modes'
	bl_description = 'Solve for the eigenmodes of the simulation setup.'

	@classmethod
	def poll(cls, context):
		"""Allow running when there are runnable tasks."""
		return (
			# Check FDTDSolverNode is Accessible
			hasattr(context, 'node')
			and hasattr(context.node, 'node_type')
			and context.node.node_type == ct.NodeType.ModeSolver
			# Check Task is Runnable
			and not context.node.solved
			and context.node.mode_solver is not None
		)

	def execute(self, context):
		"""Run all uploaded, runnable tasks."""
		node = context.node

		node.trigger_event(ct.FlowEvent.EnableLock)
		node.locked = False
		try:
			node.mode_solver.solve()
		except:  # noqa: E722
			node.trigger_event(ct.FlowEvent.DisableLock)
			self.report({'ERROR'}, 'Modal solution failed. Please check logs.')
			return {'FINISHED'}
		else:
			node.solved = True

		return {'FINISHED'}


class ReleaseSolvedModes(bpy.types.Operator):
	"""Run a modal solver on data provided to a `ModeSolverNode`."""

	bl_idname = ct.OperatorType.NodeReleaseSolvedModes
	bl_label = 'Release Solved Modes'
	bl_description = 'Release the solved eigenmodes.'

	@classmethod
	def poll(cls, context):
		"""Allow running when there are runnable tasks."""
		return (
			# Check FDTDSolverNode is Accessible
			hasattr(context, 'node')
			and hasattr(context.node, 'node_type')
			and context.node.node_type == ct.NodeType.ModeSolver
			# Check Task is Runnable
			and context.node.solved
			and context.node.mode_solver is not None
		)

	def execute(self, context):
		"""Reset the modal solver."""
		node = context.node

		node.solved = False
		node.trigger_event(ct.FlowEvent.DisableLock)
		node.mode_solver = bl_cache.Signal.InvalidateCache

		return {'FINISHED'}


class ModeSolverNode(base.MaxwellSimNode):
	"""Definition of a complete FDTD simulation, including boundary conditions, domain, sources, structures, monitors, and other configuration."""

	node_type = ct.NodeType.ModeSolver
	bl_label = 'Mode Solver'

	####################
	# - Sockets
	####################
	input_sockets: typ.ClassVar = {
		'Sim': sockets.MaxwellFDTDSimSocketDef(active_kind=FK.Value),
		'Center': sockets.ExprSocketDef(
			size=spux.NumberSize1D.Vec3,
			mathtype=MT.Real,
			physical_type=PT.Length,
			default_value=sp.ImmutableMatrix([0, 0, 0]),
		),
		'Plane Size': sockets.ExprSocketDef(
			size=spux.NumberSize1D.Vec2,
			mathtype=MT.Real,
			physical_type=PT.Length,
			default_value=sp.ImmutableMatrix([1, 1]),
		),
		'Freqs': sockets.ExprSocketDef(
			active_kind=FK.Range,
			physical_type=PT.Freq,
			default_unit=spux.THz,
			default_min=374.7406,  ## 800nm
			default_max=1498.962,  ## 200nm
			default_steps=100,
		),
		'# Modes': sockets.ExprSocketDef(
			mathtype=MT.Integer,
			default_value=3,
			abs_min=1,
		),
		'Guess neff': sockets.ExprSocketDef(
			mathtype=MT.Real,
			default_value=2.0,
		),
		## TODO: Spherical, bend, mode tracking, group index, extra PML
	}
	output_sockets: typ.ClassVar = {
		'Solved Modes': sockets.MaxwellModeSocketDef(),
		'Sim Data': sockets.MaxwellFDTDSimDataSocketDef(),
	}

	####################
	# - Properties: UI
	####################
	injection_axis: ct.SimSpaceAxis = bl_cache.BLField(ct.SimSpaceAxis.X)
	injection_direction: ct.SimAxisDir = bl_cache.BLField(ct.SimAxisDir.Plus)

	mode_pol: ModePol = bl_cache.BLField(ModePol.Arbitrary)

	solved: bool = bl_cache.BLField(False)

	####################
	# - Properties: Mode Solver
	####################
	@events.on_value_changed(
		socket_name={
			'Sim': {FK.Func, FK.Params},
			'Center': {FK.Func, FK.Params},
			'Plane Size': {FK.Func, FK.Params},
			'Freqs': {FK.Func, FK.Params},
			'# Modes': {FK.Func, FK.Params},
			'Guess neff': {FK.Func, FK.Params},
		},
	)
	def on_input_changed(self) -> None:
		"""Recomputes the mode solver in response to inputs."""
		self.mode_solver = bl_cache.Signal.InvalidateCache

	@bl_cache.cached_bl_property(
		depends_on={'injection_axis', 'injection_direction', 'mode_pol'}
	)
	def mode_solver(self) -> td.plugins.mode.ModeSolver | None:
		sim = self._compute_input('Sim', kind=FK.Value)
		center = events.realize_known(
			frozendict(
				{
					kind: self._compute_input(
						'Center', kind=kind, unit_system=ct.UNITS_TIDY3D
					)
					for kind in [FK.Func, FK.Params]
				}
			),
			freeze=True,
		)
		size_2d = events.realize_known(
			frozendict(
				{
					kind: self._compute_input(
						'Plane Size', kind=kind, unit_system=ct.UNITS_TIDY3D
					)
					for kind in [FK.Func, FK.Params]
				}
			),
			freeze=True,
		)
		freqs = events.realize_known(
			frozendict(
				{
					kind: self._compute_input(
						'Freqs', kind=kind, unit_system=ct.UNITS_TIDY3D
					)
					for kind in [FK.Func, FK.Params]
				}
			),
			freeze=True,
		)
		num_modes = events.realize_known(
			frozendict(
				{
					kind: self._compute_input('# Modes', kind=kind)
					for kind in [FK.Func, FK.Params]
				}
			),
			freeze=True,
		)
		guess_neff = events.realize_known(
			frozendict(
				{
					kind: self._compute_input('Guess neff', kind=kind)
					for kind in [FK.Func, FK.Params]
				}
			),
			freeze=True,
		)

		if not FS.check(sim) and all(
			flow is not None
			for flow in [
				center,
				size_2d,
				num_modes,
				guess_neff,
				freqs,
			]
		):
			mode_pol = self.mode_pol
			injection_direction = self.injection_direction
			_size_2d = sp.flatten(size_2d)
			size = {
				ct.SimSpaceAxis.X: (0, *_size_2d),
				ct.SimSpaceAxis.Y: (_size_2d[0], 0, _size_2d[1]),
				ct.SimSpaceAxis.Z: (*_size_2d, 0),
			}[self.injection_axis]

			return td.plugins.mode.ModeSolver(
				simulation=sim,
				plane=td.Box(center=sp.flatten(center), size=size),
				mode_spec=td.ModeSpec(
					num_modes=num_modes,
					target_neff=guess_neff,
					filter_pol=mode_pol.tidy3d_pol,
				),
				freqs=freqs,
				direction=injection_direction.plus_or_minus,
			)
		return None

	####################
	# - UI
	####################
	def draw_operators(self, _: bpy.types.Context, layout: bpy.types.UILayout):
		row = layout.row(align=True)
		row.operator(
			ct.OperatorType.NodeSolveModes,
			text='Solve',
		)
		if self.solved:
			row.operator(
				ct.OperatorType.NodeReleaseSolvedModes,
				icon='LOOP_BACK',
				text='',
			)

	def draw_props(self, _: bpy.types.Context, layout: bpy.types.UILayout):
		"""Present choices of injection axis, direction, and polarization filter."""
		row = layout.row(align=True)
		row.alignment = 'CENTER'
		row.label(text='Injection')
		layout.prop(self, self.blfields['injection_axis'], expand=True)
		layout.prop(self, self.blfields['injection_direction'], expand=True)

		row = layout.row(align=True)
		row.alignment = 'CENTER'
		row.label(text='Pol Filter')
		layout.prop(self, self.blfields['mode_pol'], text='')

	####################
	# - FlowKind.Value: Solved Modes
	####################
	@events.computes_output_socket(
		'Solved Modes',
		kind=FK.Value,
		# Loaded
		props={'mode_solver', 'solved'},
	)
	def compute_solved_modes(self, props) -> td.ModeSolverData | FS:
		mode_solver = props['mode_solver']
		solved = props['solved']
		if mode_solver is not None and solved:
			return mode_solver.data
		return FS.FlowPending

	####################
	# - FlowKind.Value: Sim Data
	####################
	@events.computes_output_socket(
		'Sim Data',
		kind=FK.Value,
		# Loaded
		props={'mode_solver', 'solved'},
	)
	def compute_sim_data_value(self, props) -> td.SimulationData | FS:
		"""Compute the particular value of the simulation domain from strictly non-symbolic inputs."""
		mode_solver = props['mode_solver']
		solved = props['solved']
		if mode_solver is not None and solved:
			return mode_solver.sim_data
		return FS.FlowPending


####################
# - Blender Registration
####################
BL_REGISTER = [
	RunModalSolve,
	ReleaseSolvedModes,
	ModeSolverNode,
]
BL_NODES = {ct.NodeType.ModeSolver: (ct.NodeCategory.MAXWELLSIM_SOLVERS)}
